     /**
      * \brief JSON Array Implementation
      *
      * The value of a JSON object can be an array. This service program deals
      * with such arrays. Arrays themselves can also contain arrays. An array
      * can only contain values and arrays but no key/value pairs.
      *
      * <br><br>
      *
      * Throughout this service program is the index always 0-based. Which
      * means that the first entry has the index 0, the second entry has the
      * index 1 and so on.
      *
      * <br><br>
      *
      * If the index is out of the bounds of the array an escape message will
      * be send. If a procedure for the retrieval of a special type of value
      * is called on an entry of another type an escape message will be send
      * too.
      *
      * \author Mihael Schmidt
      * \date   19.04.2008
      *
      * \link http://www.json.org JSON
      * \link http://www.rpgnextgen.com RPG Next Gen
      * 
      * \rev 07.03.2009 Mihael Schmidt
      *      Added <em>jsona_size</em> procedure.
      *
      * \rev 27.01.2010 Mihael Schmidt
      *      Added support for getting numbers with procedures jsona_getInt,
      *      jsona_getLong, jsona_getDouble no matter what the original type was
      *      (f. e. double value will be cast to int with jsona_getInt).
      * 
      * \rev 09.04.2011 Mihael Schmidt
      *      Fixed memory management error in toString procedure.
      * 
      * \rev 30.04.2011 Mihael Schmidt
      *      Switch from Linked List to ArrayList
      * 
      * \rev 20.07.2011 Mihael Schmidt
      *      Remove leading plus (+) from floating point numbers when creating
      *      a string from a json array.
      * 
      * \rev 23.07.2011 Mihael Schmidt
      *      Added jsona_isNull
      */

      *-------------------------------------------------------------------------
      * Copyright (c) 2007-2011 Mihael Schmidt
      * All rights reserved.
      *
      * This file is part of the JSON service program.
      *
      * JSON service program is free software: you can redistribute it and/or modify it under
      * the terms of the GNU Lesser General Public License as published by
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * JSON service program is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU Lesser General Public License for more details.
      *
      * You should have received a copy of the GNU Lesser General Public
      * License along with JSON service program.  If not, see http://www.gnu.org/licenses/.
      *-------------------------------------------------------------------------

     H nomain
     H bnddir('QC2LE')
     H copyright('Copyright (c) 2007-2011 Mihael Schmidt. All rights reserved.')


      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
     D jsona_srvpgm_init...
     D                 PR            10I 0 extproc('jsona_srvpgm_init')
      *
      /copy 'json_h.rpgle'
      /copy 'json_int_h.rpgle'
      /copy 'ceeapi_h.rpgle'
      /copy 'libc_h.rpgle'
      /copy 'arraylist_h.rpgle'


      *-------------------------------------------------------------------------
      * Constants
      *-------------------------------------------------------------------------
      /copy 'json_c.rpgle'
      /copy 'unicode_c.rpgle'


      *-------------------------------------------------------------------------
      * Global Variables
      *-------------------------------------------------------------------------
     D hex_null        S              1A   inz(x'00')
      *
      * Parameters passed to CEEDOD
      *
     D descType        S             10I 0
     D dataType        S             10I 0
     D descInfo1       S             10I 0
     D descInfo2       S             10I 0
     D length          S             10I 0
      *
     D chars           DS                  likeds(tmpl_chars)
     D chars_compact   DS                  likeds(tmpl_chars)


      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------


     /**
      * \brief Service program initialization
      *
      * String formattings are preloaded.
      */
     P jsona_srvpgm_init...
     P                 B                   export
     D                 PI            10I 0
      /free
       chars.d_quote = %char(UNICODE_DOUBLE_QUOTE) + x'00';
       chars.l_brace = ' ' + %char(UNICODE_LEFT_CURLY_BRACE) + ' ' + x'00';
       chars.r_brace = ' ' + %char(UNICODE_RIGHT_CURLY_BRACE) + ' ' + x'00';
       chars.l_bracket = ' ' + %char(UNICODE_LEFT_BRACKET) + ' ' + x'00';
       chars.r_bracket = ' ' + %char(UNICODE_RIGHT_BRACKET) + ' ' + x'00';
       chars.comma = ' ' + %char(UNICODE_COMMA) + ' ' + x'00';
       chars.colon = ' ' + %char(UNICODE_COLON) + ' ' + x'00';
       chars.true = 'true' + x'00';
       chars.false = 'false' + x'00';
       chars.null_str = 'null' + x'00';
       chars.size_d_quote = 1;
       chars.size_l_brace = 3;
       chars.size_r_brace = 3;
       chars.size_l_bracket = 3;
       chars.size_r_bracket = 3;
       chars.size_comma = 3;
       chars.size_colon = 3;
       chars.size_true = 4;
       chars.size_false = 5;
       chars.size_null_str = 4;

       chars_compact.d_quote = %char(UNICODE_DOUBLE_QUOTE) + x'00';
       chars_compact.l_brace = %char(UNICODE_LEFT_CURLY_BRACE) + x'00';
       chars_compact.r_brace = %char(UNICODE_RIGHT_CURLY_BRACE) + x'00';
       chars_compact.l_bracket = %char(UNICODE_LEFT_BRACKET) + x'00';
       chars_compact.r_bracket = %char(UNICODE_RIGHT_BRACKET) + x'00';
       chars_compact.comma = %char(UNICODE_COMMA) + x'00';
       chars_compact.colon = %char(UNICODE_COLON) + x'00';
       chars_compact.true = 'true' + x'00';
       chars_compact.false = 'false' + x'00';
       chars_compact.null_str = 'null' + x'00';
       chars_compact.size_d_quote = 1;
       chars_compact.size_l_brace = 1;
       chars_compact.size_r_brace = 1;
       chars_compact.size_l_bracket = 1;
       chars_compact.size_r_bracket = 1;
       chars_compact.size_comma = 1;
       chars_compact.size_colon = 1;
       chars_compact.size_true = 4;
       chars_compact.size_false = 5;
       chars_compact.size_null_str = 4;

       return 0;
      /end-free
     P                 E


     /**
      * \brief Create JSON array
      *
      * Creates a json array "object". Currently a linked list is used
      * as a backend for the json array.
      *
      * \author Mihael Schmidt
      * \date   19.04.2008
      *
      * \return Pointer to the JSON array
      */
     P jsona_create    B                   export
     D                 PI              *
      /free
       return arraylist_create();
      /end-free
     P                 E

     /**
      * \brief Dispose JSON array
      *
      * Disposes a JSON array and all values it contains.
      * All added arrays and objects are also disposed.
      *
      * \author Mihael Schmidt
      * \date   19.04.2008
      *
      * \param Pointer to the JSON array
      */
     P jsona_dispose   B                   export
     D                 PI
     D   array                         *
      /free
       if (array <> *null);
         jsona_clear(array);
       
         arraylist_dispose(array);
         array = *null;
       endif;
      /end-free
     P                 E


     /**
      * \brief Remove an entry from the array
      *
      * \param Pointer to the JSON array
      * \param Index of the element to be removed (0-based)
      */
     P jsona_remove    B                   export
     D                 PI                  opdesc
     D   array                         *   const
     D   index                       10I 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = arraylist_get(array : index);
       if (entry.type = JSON_TYPE_ARRAY);
         jsona_dispose(entry.array);
       elseif (entry.type = JSON_TYPE_OBJECT);
         json_dispose(entry.object);
       endif;

       arraylist_remove(array : index);
      /end-free
     P                 E


     /**
      * \brief Clear array
      *
      * Removes all elements from the array.
      *
      * \param Pointer to the JSON array
      */
     P jsona_clear     B                   export
     D                 PI
     D   array                         *   const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
     D size            S             10I 0
     D i               S             10I 0
      /free
       size = arraylist_getSize(array);
       
       for i = size-1 downto 0;
         jsona_remove(array : i);
       endfor;
      /end-free
     P                 E

     /**
      * \brief Add character entry
      *
      * Adds an entry of type character.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Value
      * \param Index (default: last)
      */
     P jsona_putString...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonArrPtr                    *   const
     D   value                    65535A   const options(*varsize)
     D   index                       10U 0 const options(*nopass)
      /free
       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpl_array_entry.type = JSON_TYPE_STRING;
       tmpl_array_entry.string = %subst(value : 1 : length);

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : length + 4);
       else;
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 
                                    length + 4 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Add boolean entry
      *
      * Adds an entry of type boolean.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Value
      * \param Index (default: last)
      */
     P jsona_putBoolean...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                         N   const
     D   index                       10U 0 const options(*nopass)
      /free
       tmpl_array_entry.type = JSON_TYPE_BOOLEAN;
       tmpl_array_entry.boolean = value;

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 5);
       else;
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 5 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Add integer entry
      *
      * Adds an entry of type integer.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Value
      * \param Index (default: last)
      */
     P jsona_putInt...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                       10I 0 const
     D   index                       10U 0 const options(*nopass)
      /free
       tmpl_array_entry.type = JSON_TYPE_INT;
       tmpl_array_entry.integer = value;

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 8);
       else;
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 8 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Add long entry
      *
      * Adds an entry of type long.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Value
      * \param Index (default: last)
      */
     P jsona_putLong...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                       20I 0 const
     D   index                       10U 0 const options(*nopass)
      /free
       tmpl_array_entry.type = JSON_TYPE_LONG;
       tmpl_array_entry.long = value;

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 12);
       else;
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 12 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Add double entry
      *
      * Adds an entry of type double.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Value
      * \param Index (default: last)
      */
     P jsona_putDouble...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                        8F   const
     D   index                       10U 0 const options(*nopass)
      /free
       tmpl_array_entry.type = JSON_TYPE_DOUBLE;
       tmpl_array_entry.double = value;

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 12);
       else;
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 12 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Add array entry
      *
      * Adds an entry of type array.
      *
      * \author Mihael Schmidt
      * \date   19.04.2008
      *
      * \param Pointer to the JSON array
      * \param Array
      * \param Index (default: last)
      */
     P jsona_putArray...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                         *   const
     D   index                       10U 0 const options(*nopass)
      *
     D entry           DS                  likeds(tmpl_array_entry)
      /free
       entry.type = JSON_TYPE_ARRAY;
       entry.array = value;

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(entry) : 32);
       else;
         arraylist_add(jsonArrPtr : %addr(entry) : 32 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Get string entry
      *
      * Returns a pointer to the value of the entry for the passed key.
      * The value is null terminated and can be processed with the %str BIF.
      *
      * <br><br>
      *
      * If the entry is not of type string an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Pointer to a null-terminated string
      */
     P jsona_getString...
     P                 B                   export
     D                 PI              *
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D tmpPtr          S               *
     D tmpDs           DS                  likeds(tmpl_array_entry )
     D                                     based(tmpPtr)
      /free
       tmpPtr = arraylist_get(jsonArrPtr : index);
       if (tmpPtr <> *null);
         if (tmpDs.type = JSON_TYPE_STRING);
           return tmpPtr + 4;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return *null; // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Get boolean entry
      *
      * Returns the boolean value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type boolean an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Boolean value of the entry
      */
     P jsona_getBoolean...
     P                 B                   export
     D                 PI              N
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D tmpPtr          S               *
     D tmpDs           DS                  likeds(tmpl_array_entry )
     D                                     based(tmpPtr)
      /free
       tmpPtr = arraylist_get(jsonArrPtr : index);
       if (tmpPtr <> *null);
         if (tmpDs.type = JSON_TYPE_BOOLEAN);
           return tmpDs.boolean;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return *off; // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Get integer entry
      *
      * Returns the integer value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type integer an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Integer value of the entry
      */
     P jsona_getInt...
     P                 B                   export
     D                 PI            10I 0
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D tmpPtr          S               *
     D tmpDs           DS                  likeds(tmpl_array_entry )
     D                                     based(tmpPtr)
      /free
       tmpPtr = arraylist_get(jsonArrPtr : index);
       if (tmpPtr <> *null);

         if (tmpDs.type = JSON_TYPE_INT);
           return tmpDs.integer;
         elseif (tmpDs.type = JSON_TYPE_LONG);
           monitor;
             return tmpDs.long;
             on-error *all;
               sendEscapeMessage(MSG_TYPE_CAST_EXCEPTION);
           endmon;
         elseif (tmpDs.type = JSON_TYPE_DOUBLE);
           monitor;
             return %int(tmpDs.double);
             on-error *all;
               sendEscapeMessage(MSG_TYPE_CAST_EXCEPTION);
           endmon;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;

       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return -1; // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Get long entry
      *
      * Returns the long value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type long an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Long value of the entry
      */
     P jsona_getLong...
     P                 B                   export
     D                 PI            20I 0
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D tmpPtr          S               *
     D tmpDs           DS                  likeds(tmpl_array_entry )
     D                                     based(tmpPtr)
      /free
       tmpPtr = arraylist_get(jsonArrPtr : index);
       if (tmpPtr <> *null);

         if (tmpDs.type = JSON_TYPE_LONG);
           return tmpDs.long;
         elseif (tmpDs.type = JSON_TYPE_INT);
           return tmpDs.integer;
         elseif (tmpDs.type = JSON_TYPE_DOUBLE);
           monitor;
             return %int(tmpDs.double);
             on-error *all;
               sendEscapeMessage(MSG_TYPE_CAST_EXCEPTION);
           endmon;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;

       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return -1; // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Get double entry
      *
      * Returns the double value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type double an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Double value of the entry
      */
     P jsona_getDouble...
     P                 B                   export
     D                 PI             8F
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D tmpPtr          S               *
     D tmpDs           DS                  likeds(tmpl_array_entry )
     D                                     based(tmpPtr)
      /free
       tmpPtr = arraylist_get(jsonArrPtr : index);
       if (tmpPtr <> *null);

         if (tmpDs.type = JSON_TYPE_DOUBLE);
           return tmpDs.double;
         elseif (tmpDs.type = JSON_TYPE_INT);
           return %float(tmpDs.integer);
         elseif (tmpDs.type = JSON_TYPE_LONG);
           monitor;
             return %float(tmpDs.long);
             on-error *all;
               sendEscapeMessage(MSG_TYPE_CAST_EXCEPTION);
           endmon;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;

       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return -1.0; // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Get array entry
      *
      * Returns the array value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type array an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Pointer to the JSON array
      */
     P jsona_getArray...
     P                 B                   export
     D                 PI              *
     D   array                         *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = arraylist_get(array : index);
       if (ptr <> *null);
         if (entry.type = JSON_TYPE_ARRAY);
           return entry.array;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return *null; // dummy => will never come to this point
      /end-free
     P                 E



     /**
      * \brief String representation of the JSON array in JSON syntax
      *
      * Creates a null terminated string representation of the JSON array in
      * JSON syntax.
      *
      * <br><br>
      *
      * For entries of type string double quotes and backslashes will be escaped
      * (" => \" and \ => \\).
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON array
      * 
      * \return Pointer to the JSON string for the passed array
      * 
      * \info The allocated memory of the returned string must be deallocated
      *       by the caller.
      */
     P jsona_toString...
     P                 B                   export
     D                 PI              *
     D   jsonArrPtr                    *   const
     D   pOutputFormat...
     D                               10I 0 const options(*nopass)
     D   pPrettyPrintFormat...
     D                                     likeds(json_prettyPrintFormat)
     D                                     const options(*nopass)
     D   pIndentationLevel...
     D                               10I 0 options(*nopass)
      *
     D outputFormat...
     D                 S             10I 0
     D prettyPrintOption...
     D                 S               N   inz(*off)
     D prettyPrintFormat...
     D                 DS                  likeds(json_prettyPrintFormat) inz
     D indentationLevel...
     D                 S             10I 0 inz(0)
      *
     D l_chars         DS                  likeds(tmpl_chars) inz
      *
     D json_string_ptr...
     D                 S               *
     D json_string_length...
     D                 S             10I 0
     D tokenList       S               *
     D tmpString       S            310A
     D tmpLength       S             10I 0
     D entry           DS                  likeds(tmpl_array_entry)
     D                                     based(entryPtr)
     D ptr             S               *
     D x               S             10I 0
     D newline         S              2A   inz(x'0D25')
     D i               S             10I 0
     D size            S             10I 0
      /free
       if (%parms() >= 2);
         outputFormat = pOutputFormat;
       endif;

       if (%parms() >= 3);
         prettyPrintFormat = pPrettyPrintFormat;
       endif;

       if (%parms() = 4);
         indentationLevel = pIndentationLevel;
       endif;

       exsr init;

       if (outputFormat = JSON_OUTPUT_PRETTY_PRINT);
         prettyPrintOption = *on;
       endif;

       // initialization
       //  - write start tag
       //  - init token list
       tokenList = arraylist_create();

       if (prettyPrintOption and prettyPrintFormat.breakBeforeArrayBrace);
         arraylist_add(tokenList : %addr(newline) : %size(newline));
         prettyPrint(tokenList : prettyPrintFormat : indentationLevel);
       endif;

       arraylist_add(tokenList : %addr(l_chars.l_bracket) : 
                     l_chars.size_l_bracket);

       if (prettyPrintOption);
         prettyPrint(tokenList : prettyPrintFormat : indentationLevel);

         if (prettyPrintFormat.breakAfterArrayBrace);
           indentationLevel += 1;
           arraylist_add(tokenList : %addr(newline) : %size(newline));
         endif;

       endif;


       // process entries
       size = arraylist_getSize(jsonArrPtr) - 1;
       for i = 0 to size;
         entryPtr = arraylist_get(jsonArrPtr : i);

         if (prettyPrintOption);
           prettyPrint(tokenList : prettyPrintFormat : indentationLevel);
         endif;

         select;
           when (entry.type = JSON_TYPE_STRING);
             arraylist_add(tokenList : %addr(l_chars.d_quote) : 
                           l_chars.size_d_quote);
             arraylist_add(tokenList : %addr(entry.string) :
                           strlen(%addr(entry.string)));
             escapeLastToken(tokenList);
             arraylist_add(tokenList : %addr(l_chars.d_quote) : 
                           l_chars.size_d_quote);

           when (entry.type = JSON_TYPE_INT);
             tmpString = %char(entry.integer);
             tmpLength = %len(%trimr(tmpString));
             arraylist_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = JSON_TYPE_LONG);
             tmpString = %char(entry.long);
             tmpLength = %len(%trimr(tmpString));
             arraylist_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = JSON_TYPE_DOUBLE);
             tmpString = %triml(%char(entry.double) : '+');
             tmpLength = %len(%trimr(tmpString));
             // JSON syntax expects a . as delimiter (and not ,)
             tmpString = %xlate(',' : '.' : tmpString);
             arraylist_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = JSON_TYPE_BOOLEAN);
             if (entry.boolean);
               arraylist_add(tokenList : %addr(l_chars.true):l_chars.size_true);
             else;
               arraylist_add(tokenList : %addr(l_chars.false) : 
                             l_chars.size_false);
             endif;

           when (entry.type = JSON_TYPE_ARRAY);
             ptr = jsona_toString(entry.array : outputFormat :
                                  prettyPrintFormat : indentationLevel);
             arraylist_add(tokenList : ptr : strlen(ptr));
             dealloc(n) ptr;

           when (entry.type = JSON_TYPE_NULL);
             arraylist_add(tokenList : %addr(entry.string) : 4);

           when (entry.type = JSON_TYPE_OBJECT);
             ptr = json_toString(entry.object : outputFormat :
                                 prettyPrintFormat : indentationLevel);
             arraylist_add(tokenList : ptr : strlen(ptr));

         endsl;


         // add comma if it is not the last entry
         if (entryPtr <> arraylist_getLast(jsonArrPtr));
           arraylist_add(tokenList : %addr(l_chars.comma) : l_chars.size_comma);

           if (prettyPrintOption and prettyPrintFormat.breakAfterElement);
             arraylist_add(tokenList : %addr(newline) : %size(newline));
           endif;

         endif;

       endfor;

       if (prettyPrintOption);
         if (prettyPrintFormat.breakAfterArrayBrace);
           arraylist_add(tokenList : %addr(newline) : %size(newline));
           indentationLevel -= 1;
         endif;
         prettyPrint(tokenList : prettyPrintFormat : indentationLevel);
       endif;

       // write end tag
       arraylist_add(tokenList : %addr(l_chars.r_bracket) : 
                     l_chars.size_r_bracket);

       // calculate json string length
       size = arraylist_getSize(tokenList) - 1;
       for i = 0 to size;
         ptr = arraylist_get(tokenList : i);
         json_string_length += strlen(ptr);
       endfor;
       // +1 for hex Null
       json_string_length += 1;

       // allocate memory for json string
       json_string_ptr = %alloc(json_string_length);

       // build json string
       size = arraylist_getSize(tokenList) - 1;
       for i = 0 to size;
         ptr = arraylist_get(tokenList : i);
         tmpLength = strlen(ptr);
         memcpy(json_string_ptr + x : ptr : tmpLength);
         x += tmpLength;
       endfor;

       // append hex null
       memcpy(json_string_ptr + x : %addr(hex_null) : 1);

       // clean up
       if (tokenList <> *null);
         arraylist_dispose(tokenList);
       endif;

       return json_string_ptr;

       begsr init;
         if (outputFormat = JSON_OUTPUT_COMPACT);
           l_chars = chars_compact;
         else;
           l_chars = chars;
         endif;
       endsr;
      /end-free
     P                 E


     /**
      * \brief Add null value entry
      *
      * Adds an entry of type null to the array.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Index (default: last)
      */
     P jsona_putNull   B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const options(*nopass)
      /free
       tmpl_array_entry.type = JSON_TYPE_NULL;
       tmpl_array_entry.string = 'null';

       if (%parms() = 1);
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 4 + 4);
       else;
         arraylist_add(jsonArrPtr : %addr(tmpl_array_entry ) : 4 + 4 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Add JSON object entry
      *
      * Adds an entry of type object to the array.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Pointer to the JSON object
      * \param Index (default: last)
      */
     P jsona_putObject...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   objPtr                        *   const
     D   index                       10U 0 const options(*nopass)
      *
     D entry           DS                  likeds(tmpl_array_entry)
      /free
       entry.type = JSON_TYPE_OBJECT;
       entry.object = objPtr;

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(entry) : 32);
       else;
         arraylist_add(jsonArrPtr : %addr(entry) : 32 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Get object entry
      *
      * Returns a pointer to the object.
      *
      * <br><br>
      *
      * If the entry is not of type object an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Pointer to the JSON object
      */
     P jsona_getObject...
     P                 B                   export
     D                 PI              *
     D   array                         *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = arraylist_get(array : index);
       if (ptr <> *null);
         if (entry.type = JSON_TYPE_OBJECT);
           return entry.object;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return *null; // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Get array size
      *
      * Returns the number of elements in the array.
      *
      * \param Pointer to the json array
      *
      * \return number of elements
      */
     P jsona_size      B                   export
     D                 PI            10U 0
     D   jsonArrPtr                    *   const
      /free
       return arraylist_getSize(jsonArrPtr);
      /end-free
     P                 E


     /**
      * \brief Get entry type
      *
      * Returns the type of the entry.
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Entry type
      *
      * \throws CPF9898 Array element not found
      */
     P jsona_getEntryType...
     P                 B                   export
     D                 PI            10I 0
     D   array                         *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = arraylist_get(array : index);
       if (ptr <> *null);
         return entry.type;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
         return 0; // dummy
       endif;
      /end-free
     P                 E


     /**
      * \brief Checks if entry is of special type null
      * 
      * \param Pointer to the JSON object
      * \param Key
      * 
      * \return *on = entry is of type null <br>
      *         *off = entry is of another type
      */
     P jsona_isNull...
     P                 B                   export
     D                 PI              N   
     D   array                         *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = arraylist_get(array : index);
       if (ptr <> *null);
         return (entry.type = JSON_TYPE_NULL);
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
         return *off; // dummy
       endif;
      /end-free
     P                 E
     